/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: ConnectionManager.java,v 1.35.2.1 2004/09/07 07:53:52 durieuxp Exp $
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.dbm;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import javax.naming.NamingException;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.StringRefAddr;
import javax.sql.ConnectionEvent;
import javax.sql.ConnectionEventListener;
import javax.sql.DataSource;
import javax.sql.XAConnection;
import javax.transaction.RollbackException;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import javax.transaction.TransactionManager;
import javax.transaction.xa.XAResource;

import org.objectweb.jonas.common.Log;
import org.objectweb.jonas.jdbc_xa.XADataSourceImpl;
import org.objectweb.jonas.jtm.TransactionService;
import org.objectweb.jonas.management.ReconfigDispatcher;
import org.objectweb.jonas.management.ReconfiguredProp;
import org.objectweb.jonas.service.ServiceManager;
import org.objectweb.transaction.api.ResourceManagerEvent;
import org.objectweb.transaction.api.ResourceManagerEventListener;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;

/**
 * JOnAS generic implementation of DataSource to manage an underlying
 * JDBC-XA driver (XADataSource, XAConnection, XAResource, ...)
 * This class acts as a factory for Connection objects.
 * Its main goal is to manage a pool of XAConnection objects, to avoid 
 * closing the actual Connection on DBMS, and to deal with distributed
 * transactions.
 * This DataSource is registered in JNDI (=> implements Referenceable)
 *
 * @author Philippe Durieux
 * <p>
 * Contributor(s): 
 * <p>
 * 00/01/08 Markus Fritz make connection pooling thread safe (with same tx)
 *          introduce MaxAge for connections
 *          optimize pooling
 *          add some 'paranoia code': test conns before returning them from the pool.
 * <p>
 * 00/18/04 Jun Inamori (j-office@osa.att.ne.jp)
 *          For closing correctly all connections at server shutdown.
 * <p>01/11/06 Christophe Ney cney@batisseurs.com for Lutris Technologies
 *          Added ResourceManagerListener mechanism.
 * <p>02/01/15 Dean Jennings - Map   instead of Hashtable  for cmList
 * <p>Adriana Danes
 * <p>03/01/13 - Manage datasource name additionally to the jndi name for the datasource using this ConnectionManager
 * <p>03/04/08 - ConnectionManager extends ReconfigDispatcher in order to send reconfiguration Notifications to ReconfigManager
 * <p>03/05/23 - Introduce pool size configuration
 */
public class ConnectionManager extends ReconfigDispatcher 
    implements DataSource, 
               Referenceable, 
               Serializable,
               ConnectionEventListener,
               ConnectionManagerMBean {

    private static Logger logger = null;

    // The static table of all the ConnectionManager objects
    // Each ResourceManager has its own ConnectionManager (i.e. DataSource)
    private static Map cmList = new HashMap();

    /**
     * TransactionManager object
     * @serial
     */
    private TransactionManager tm = null;

    /**
     * underlaying XA DataSource
     * @serial
     */
    private XADataSourceImpl xads = null;

    /**
     * The pool of XAConnections managed by this Connection Manager
     * @serial
     */
    private Pool pool = null;

    /**
     * @return The pool associated to this datasource
     */
    public Pool getPool() {
        return pool;
    }

    // ----------------------------------------------------------
    // properties
    // ----------------------------------------------------------

    /**
     * @serial for JNDI
     */
    private String dSName = null;

    /**
     * @serial datasource name
     */
    private String datasourceName;

    /**
     * @serial url for database
     */
    private String url = null;
    /**
     * @serial JDBC driver Class
     */
    private String className = null;
    /**
     * @serial default user
     */
    private String userName = null;
    /**
     * @serial default passwd
     */
    private String password = null;

    /**
     * @serial 
     */
    private int isolationLevel = -1;
    private String isolationStr = null;

    /**
     * @serial 
     */
    private String currentMapperName = null;

    /**
     * Name of the client (client case)
     * TODO : must be removed. A client can lookup a datasource
     */
    private boolean isClient = false;

    /**
     * MBean variable
     */
    private String xaName = null;
    private String dsDescription = null;

    private static ResourceManagerEventListener rmListener = null;

    /**
     * Value used as sequence number by reconfiguration notifications
     */
    long sequenceNumber = 0;

    long getSequenceNumber() {
        return ++sequenceNumber;
    }
    
    // -----------------------------------------------------------------
    // Constructors
    // -----------------------------------------------------------------

    /**
     * Constructor for Factory
     */
    public ConnectionManager() throws Exception {
        logger = Log.getLogger(Log.JONAS_DBM_PREFIX);
        super.initLogger(Log.getLogger(Log.JONAS_MANAGEMENT_PREFIX));

        // Set the TransactionManager
        try {
            // Modified for use with the gcj compiler
            ServiceManager sm = ServiceManager.getInstance();
            TransactionService ts = (TransactionService) sm.getTransactionService();
            tm = ts.getTransactionManager();
        } catch (Exception e) {
            throw e;
        }

        // set the Transaction Manager as being the listener for this RM
        setRMEListener((ResourceManagerEventListener) tm);       

        // Create the underlaying JDBC-XA driver
        // We do not use JNDI here. (should we do ?)
        // AD : do this before the Pool creation in order to allow the Pool constructor to access xads
        xads = new XADataSourceImpl();
        pool = new Pool(this, xads);
    }

    /**
     * Constructor for the client case
     * @param isClient true if this is the client case
     * TODO : must be removed. A client can lookup a datasource
     */
    public ConnectionManager(boolean isClient) throws Exception { 
        logger = Log.getLogger(Log.JONAS_DBM_PREFIX);
        super.initLogger(Log.getLogger(Log.JONAS_MANAGEMENT_PREFIX));
        this.isClient = isClient;
        xads = new XADataSourceImpl();
    }

    /**
     * This manager is in the client case or not ?
     * @return boolean true if this is the client case
     * TODO : must be removed. A client can lookup a datasource
     */
    public boolean isClientCase() {
        return isClient;
    }

    // -----------------------------------------------------------------
    // Properties
    // -----------------------------------------------------------------

    public void setRMEListener(ResourceManagerEventListener rmel) {
        rmListener = rmel;
    }
  
    /**
     * @param s jndi name for the datasource 
     */
    public void setDSName(String s) {
        dSName = s;
        xads.setDataSourceName(s); // for getRMID!
        // Add it to the list
        cmList.put(s, this);
    }

    public String getUrl() {
        return url;
    }
    public void setUrl(String s) {
        url = s;
        // Forward it to JDBC-XA driver
        xads.setUrl(s);
    }

    public String getClassName() {
        return className;
    }
    public void setClassName(String s) throws ClassNotFoundException {
        className = s;
        // Forward it to JDBC-XA driver
        xads.setClassName(s);
    }

    public String getUserName() {
        return userName;
    }
    public void setUserName(String s) {
        userName = s;
        // Forward it to JDBC-XA driver
        xads.setUserName(s);
    }

    public String getPassword() {
        return password;
    }
    public void setPassword(String s) {
        password = s;
        // Forward it to JDBC-XA driver
        xads.setPassword(s);
    }

    public void setTransactionIsolation(String level) {
        logger.log(BasicLevel.DEBUG, level);
        if (level.equals("serializable")) {
            isolationLevel = Connection.TRANSACTION_SERIALIZABLE;
        } else if (level.equals("none")) {
            isolationLevel = Connection.TRANSACTION_NONE;
        } else if (level.equals("read_committed")) {
            isolationLevel = Connection.TRANSACTION_READ_COMMITTED;
        } else if (level.equals("read_uncommitted")) {
            isolationLevel = Connection.TRANSACTION_READ_UNCOMMITTED;
        } else if (level.equals("repeatable_read")) {
            isolationLevel = Connection.TRANSACTION_REPEATABLE_READ;
        } else {
            isolationStr = "default";
            return;
        }
        // Forward it to JDBC-XA driver
        isolationStr = level;
        xads.setTransactionIsolation(isolationLevel);
    }

    public String getTransactionIsolation() {
        logger.log(BasicLevel.DEBUG, isolationStr);
        return isolationStr;
    }

    public void setMapperName(String mappername) {
        currentMapperName = mappername;
    }

    public String getMapperName() {
        return currentMapperName;
    }

    /**
     * Configure the Connection pool. Called by the Container at init.
     * Configuration can be set in datasource.properties files.
     * @param connchecklevel JDBC connection checking level
     * @param connmaxage JDBC connection maximum age
     * @param maxopentime JDBC connection maximum open time
     * @param connteststmt SQL query for test statement
     * @param minconpool Min size for the connection pool
     * @param maxconpool Max size for the connection pool
     * @param maxwaittime Max time to wait for a connection (in seconds)
     * @param maxwaiters Max nb of waiters for a connection
     * @param samplingperiod sampling period in sec.
     * @throw ServiceException if could not create the initial items in the pool
     */
    public void poolConfigure(String connchecklevel,
                              String connmaxage,
                              String maxopentime,
                              String connteststmt,
                              String minconpool,
                              String maxconpool,
                              String maxwaittime,
                              String maxwaiters,
                              String samplingperiod) {

        // Configure pool
        pool.setCheckLevel((new Integer(connchecklevel)).intValue());
        // set con max age BEFORE min/max pool size.
        pool.setMaxAge((new Integer(connmaxage)).intValue());
        pool.setMaxOpenTime((new Integer(maxopentime)).intValue());
        pool.setTestStatement(connteststmt);
        pool.setPoolMin((new Integer(minconpool)).intValue());
        pool.setPoolMax((new Integer(maxconpool)).intValue());
        pool.setMaxWaitTime((new Integer(maxwaittime)).intValue());
        pool.setMaxWaiters((new Integer(maxwaiters)).intValue());
        pool.setSamplingPeriod((new Integer(samplingperiod)).intValue());
        if (logger.isLoggable(BasicLevel.DEBUG)) {
            logger.log(BasicLevel.DEBUG, "ConnectionManager configured with:");
            logger.log(BasicLevel.DEBUG, "   jdbcConnCheckLevel  = " + connchecklevel);
            logger.log(BasicLevel.DEBUG, "   jdbcConnMaxAge      = " + connmaxage);
            logger.log(BasicLevel.DEBUG, "   jdbcMaxOpenTime     = " + maxopentime);
            logger.log(BasicLevel.DEBUG, "   jdbcTestStmt        = " + connteststmt);
            logger.log(BasicLevel.DEBUG, "   minConPool          = " + pool.getPoolMin());
            logger.log(BasicLevel.DEBUG, "   maxConPool          = " + pool.getPoolMax());
            logger.log(BasicLevel.DEBUG, "   maxWaitTime         = " + pool.getMaxWaitTime());
            logger.log(BasicLevel.DEBUG, "   maxWaiters          = " + pool.getMaxWaiters());
            logger.log(BasicLevel.DEBUG, "   samplingPeriod      = " + pool.getSamplingPeriod());
        }
    }

    // -----------------------------------------------------------------
    // DataSource implementation
    // -----------------------------------------------------------------

    /**
     * Attempt to establish a database connection.
     *
     * @return a Connection to the database
     *
     * @exception java.sql.SQLException - if a database-access error occurs.
     */
    public Connection getConnection() throws SQLException {
        return getConnection(userName, password);
    }

    /**
     * Attempt to establish a database connection.
     *
     * @param user - the database user on whose behalf the Connection is being made
     * @param password - the user's password
     *
     * @return a Connection to the database
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public Connection getConnection(String username, String password) throws SQLException {
        XAConnection xac = null;
        Transaction tx = null;
        try {
            tx = tm.getTransaction();
        } catch (NullPointerException n) {
            // current is null: we are not in a JOnAS Server.
            logger.log(BasicLevel.ERROR, "ConnectionManager: should not be used outside a JOnAS Server");
        } catch (SystemException e) {
            logger.log(BasicLevel.ERROR, "ConnectionManager: getTransaction failed:" + e);
        }
        logger.log(BasicLevel.DEBUG, "Tx = " + tx);

        // Check if available XAConnection in the pool for this user
        PoolItem pi = pool.openConnection(username, (Transaction) tx);
        // get the xac associated to the new PoolItem
        xac = pi.getXACon();

        Connection ret = xac.getConnection();
        // Enlist XAResource if we are actually in a transaction
        if (tx != null) {
            if (pi.getOpenCount() == 1) { // Only if first/only thread
                try {
                    if (logger.isLoggable(BasicLevel.DEBUG)) {
                        logger.log(BasicLevel.DEBUG, "enlist XAResource on " + tx);
                    }
                    tx.enlistResource(xac.getXAResource());
                } catch (RollbackException e) {
                    // Although tx has been marked to be rolled back, 
                    // XAResource has been correctly enlisted.
                    logger.log(BasicLevel.WARN, "XAResource enlisted, but tx is marked rollback");
                } catch (Exception e) {
                    logger.log(BasicLevel.ERROR, "Cannot enlist XAResource:" + e);
                    logger.log(BasicLevel.ERROR, "Connection will not be enlisted in a transaction");
                    // should return connection in the pool XXX
                    throw new SQLException("Cannot enlist XAResource");
                }
                // We are inside a transaction: no autocommit!
                ret.setAutoCommit(false);
            }
        } else {
            ret.setAutoCommit(true);    // in case we do not start a Tx
        }
        // We are not in a transaction yet: we just make a pre-registration.
        // We are in a stateful bean, put it in the list too.
        if (!pi.isRME()) {
            logger.log(BasicLevel.DEBUG, "register this connection to the TM.");
            pi.setRME(true);
            rmListener.connectionOpened(pi);
        }
        // return a Connection object
        return ret;
    }

    /**
     * Get the log writer for this data source. 
     * 
     * The log writer is a character output stream to which all logging and 
     * tracing messages for this data source object instance will be printed. 
     * This includes messages printed by the methods of this object, messages 
     * printed by methods of other objects manufactured by this object, and so on.
     * Messages printed to a data source specificlog writer are not printed to the 
     * log writer associated with the Drivermanager class. When a data
     * source object is created the log writer is initially null, in other words, 
     * logging is disabled.
     *
     * @return the log writer for this data source, null if disabled
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public java.io.PrintWriter getLogWriter() throws SQLException {
        return xads.getLogWriter();
    }


    /**
     * Set the log writer for this data source. 
     * 
     * The log writer is a character output stream to which all logging and 
     * tracing messages for this data source object instance will be printed. 
     * This includes messages printed by the methods of this object, messages printed
     * by methods of other objects manufactured by this object, and so on. 
     * Messages printed to a data source specific log writer are not printed to 
     * the log writer associated with the Drivermanager class. When a data
     * source object is created the log writer is initially null, in other words, 
     * logging is disabled.
     * 
     * @param out - the new log writer; to disable, set to null
     *
     * @exception SQLException - if a database-access error occurs.
     */
    public void setLogWriter(java.io.PrintWriter out) throws SQLException {
        xads.setLogWriter(out);
    }


    /**
     * Sets the maximum time in seconds that this data source will wait while 
     * attempting to connect to a database.
     * A value of zero specifies that the timeout is the default system timeout 
     * if there is one; 
     * otherwise it specifies that there is no timeout. 
     * When a data source object is created the login timeout is initially zero.
     *
     * @param seconds - the data source login time limit
     *
     * @exception SQLException - if a database access error occurs.
     */
    public void setLoginTimeout(int seconds) throws SQLException {
        xads.setLoginTimeout(seconds);
    }

    /**
     * Gets the maximum time in seconds that this data source can wait while attempting 
     * to connect to a database.
     * A value of zero means that the timeout is the default system timeout if there is one; 
     * otherwise it means that there is no timeout. When a data source object is created 
     * the login timeout is initially zero.
     *
     * @return the data source login time limit
     *
     * @exception SQLException - if a database access error occurs.
     */
    public int getLoginTimeout() throws SQLException {
        return xads.getLoginTimeout();
    }

    // -----------------------------------------------------------------
    // ConnectionEventListener implementation
    // -----------------------------------------------------------------

    /**
     * Invoked when the application calls close() on its representation 
     * of the connection.
     * @param event - an event object describing the source of the event
     */
    public void connectionClosed(ConnectionEvent event) {
        logger.log(BasicLevel.DEBUG, " ");

        XAConnection xac = (XAConnection) event.getSource();
        PoolItem pi = pool.closeConnection(xac, XAResource.TMSUCCESS);

        // remove it from the list of open connections for this thread
        // only if it was opened outside a tx.
        if (pi != null && pi.isRME()) {
            logger.log(BasicLevel.DEBUG, "unregister this connection to the TM.");
            pi.setRME(false);
            rmListener.connectionClosed(pi);
        }
    }

    /**
     * Invoked when a fatal connection error occurs, just before an
     * SQLException is thrown to the application.
     * @param event - an event object describing the source of the event
     */
    public void connectionErrorOccurred(ConnectionEvent event) {
        logger.log(BasicLevel.DEBUG, "");

        XAConnection xac = (XAConnection) event.getSource();
        PoolItem pi = pool.closeConnection(xac, XAResource.TMFAIL);

        // remove it from the list of open connections for this thread
        // only if it was opened outside a tx.
        if (pi != null && pi.isRME()) {
            logger.log(BasicLevel.DEBUG, "unregister this connection to the TM.");
            pi.setRME(false);
            rmListener.connectionErrorOccured(pi);
        }
    }

    // -----------------------------------------------------------------
    // Referenceable implementation
    // -----------------------------------------------------------------

    /**
     * Retrieves the Reference of this object. Used at binding time by JNDI
     * to build a reference on this object.
     * 
     * @return  The non-null Reference of this object. 
     * @exception  NamingException  If a naming exception was encountered while 
     * retrieving the reference.  
     */
    public Reference getReference() throws NamingException {

        Reference ref = new Reference (this.getClass().getName(), 
                                       "org.objectweb.jonas.dbm.DataSourceFactory",
                                       null);
        // These values are used by ObjectFactory (see DataSourceFactory.java)
        ref.add(new StringRefAddr("datasource.name", getDSName()));
        ref.add(new StringRefAddr("datasource.url", getUrl()));
        ref.add(new StringRefAddr("datasource.classname", getClassName()));
        ref.add(new StringRefAddr("datasource.username", getUserName()));
        ref.add(new StringRefAddr("datasource.password", getPassword()));
        ref.add(new StringRefAddr("datasource.isolationlevel", getTransactionIsolation()));
        ref.add(new StringRefAddr("datasource.mapper", getMapperName()));
        Integer checklevel = new Integer(pool.getCheckLevel());
        ref.add(new StringRefAddr("connchecklevel", checklevel.toString()));
        Integer maxage = new Integer(pool.getMaxAge());
        ref.add(new StringRefAddr("connmaxage", maxage.toString()));
        Integer maxopentime = new Integer(pool.getMaxOpenTime());
        ref.add(new StringRefAddr("maxopentime", maxopentime.toString()));
        ref.add(new StringRefAddr("connteststmt", pool.getTestStatement()));
        Integer minpool = new Integer(pool.getPoolMin());
        ref.add(new StringRefAddr("minconpool", minpool.toString()));
        Integer maxpool = new Integer(pool.getPoolMax());
        ref.add(new StringRefAddr("maxconpool", maxpool.toString()));
        Integer maxwaittime = new Integer(pool.getMaxWaitTime());
        ref.add(new StringRefAddr("maxwaittime", maxwaittime.toString()));
        Integer maxwaiters = new Integer(pool.getMaxWaiters());
        ref.add(new StringRefAddr("maxwaiters", maxwaiters.toString()));
        Integer samplingperiod = new Integer(pool.getSamplingPeriod());
        ref.add(new StringRefAddr("samplingperiod", samplingperiod.toString()));
        return ref;
    }

    /**
     * get the ConnectionManager matching the DataSource name
     */
    public static ConnectionManager getConnectionManager(String dsname) {
       
        ConnectionManager cm = (ConnectionManager) cmList.get(dsname);
        return cm;
    }

    public void closeAllConnection() {
        logger.log(BasicLevel.DEBUG, "");
        pool.closeAllConnections();
    }

    public static ResourceManagerEventListener getResourceManagerEventListener() {
        return rmListener;
    }

    /**
     * MBean method:
     * @return Integer number of open connection in this ds
     */
    public Integer getCurrentOpened() {
        return new Integer(pool.getCurrentOpened());
    }

    /**
     * MBean method:
     * @return Integer number of busy connection in this ds
     */
    public Integer getCurrentBusy() {
        return new Integer(pool.getCurrentBusy());
    }

    /**
     * MBean method:
     * @return Integer number of connection served
     */
    public Integer getOpenedCount() {
        return new Integer(pool.getOpenedCount());
    }

    /**
     * MBean method:
     * @return Integer number of connection failures
     */
    public Integer getConnectionFailures() {
        return new Integer(pool.getConnectionFailures());
    }

    /**
     * MBean method:
     * @return Integer number of connection leaks
     */
    public Integer getConnectionLeaks() {
        return new Integer(pool.getConnectionLeaks());
    }

    /**
     * Management method:
     * @return Integer number of xa connection served
     */
    public Integer getServedOpen() {
        return new Integer(pool.getServedOpen());
    }

    /**
     * Management method:
     * @return Integer number of open calls that were rejected
     */
    public Integer getRejectedOpen() {
        return new Integer(pool.getRejectedOpen());
    }

    /**
     * Management method:
     * @return Integer number of open calls that were rejected
     */
    public Integer getRejectedFull() {
        return new Integer(pool.getRejectedFull());
    }

    /**
     * Management method:
     * @return Integer number of open calls that were rejected
     */
    public Integer getRejectedTimeout() {
        return new Integer(pool.getRejectedTimeout());
    }

    /**
     * Management method:
     * @return Integer number of open calls that were rejected
     */
    public Integer getRejectedOther() {
        return new Integer(pool.getRejectedOther());
    }

    /**
     * MBean method:
     * @return Integer current number of connection waiters
     */
    public Integer getCurrentWaiters() {
        return new Integer(pool.getCurrentWaiters());
    }

    /**
     * MBean method:
     * @return Maximum nb of waiters since datasource creation.
     */
    public Integer getWaitersHigh() {
        return new Integer(pool.getWaitersHigh());
    }

    /**
     * MBean method:
     * @return Maximum nb of waiters in last sampling period
     */
    public Integer getWaitersHighRecent() {
        return new Integer(pool.getWaitersHighRecent());
    }

    /**
     * MBean method:
     * @return Maximum nb of busy connections in last sampling period
     */
    public Integer getBusyMaxRecent() {
        return new Integer(pool.getBusyMaxRecent());
    }

    /**
     * MBean method:
     * @return Minimum nb of busy connections in last sampling period
     */
    public Integer getBusyMinRecent() {
        return new Integer(pool.getBusyMinRecent());
    }

    /**
     * MBean method:
     * @return Total nb of waiters since datasource creation.
     */
    public Integer getWaiterCount() {
        return new Integer(pool.getWaiterCount());
    }

    /**
     * MBean method:
     * @return Total waiting time (millisec) since datasource creation.
     */
    public Long getWaitingTime() {
        // "return new Long(pool.getWaitingTime())" doesn't work!?
        long l = pool.getWaitingTime();
        return new Long(l);
    }

    /**
     * MBean method:
     * @return Maximum waiting time (millisec) since datasource creation.
     */
    public Long getWaitingHigh() {
        long l = pool.getWaitingHigh();
        return new Long(l);
    }

    /**
     * MBean method:
     * @return Maximum waiting time (millisec) in last sampling period
     */
    public Long getWaitingHighRecent() {
        long l = pool.getWaitingHighRecent();
        return new Long(l);
    }

    /**
     * MBean method:
     * @return number of connection in tx
     */
    public Integer getCurrentInTx() {
        return new Integer(pool.getCurrentInTx());
    }

    /**
     * MBean method:
     * @return Jndi name of the datasource
     */
    public String getDSName() {
        return dSName;
    }

    /**
     * @return datasource name
     */
    public String getDatasourceName() {
        return datasourceName;
    }

    /**
     * @param s datasource name
     */
    public void setDatasourceName(String s) {
        datasourceName = s;
    }

    /**
     * MBean method:
     * @return connection level
     */
    public Integer getJDBCConnCheckLevel() {
        return new Integer(pool.getCheckLevel());
    }

    /**
     * MBean method allowing to set the JDBC connection checking level value
     * @param level connection level
     */
    public void setJDBCConnCheckLevel(Integer level) {
        pool.setCheckLevel(level.intValue());
        String propName = DataBaseServiceImpl.CONNCHECKLEVEL;
        ReconfiguredProp prop = new ReconfiguredProp(propName, level.toString());
        // Send a reconfiguration notification to the listener MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * MBean method:
     * @return connection max age
     */
    public Integer getJDBCConnMaxAge() {
        return new Integer(pool.getMaxAge());
    }

    /**
     * MBean method allowing to set the JDBC connection maximum age
     * @param mn connection max age
     */
    public void setJDBCConnMaxAge(Integer mn) {
        pool.setMaxAge(mn.intValue());
        String propName = DataBaseServiceImpl.CONNMAXAGE;
        ReconfiguredProp prop = new ReconfiguredProp(propName, mn.toString());
        // Send a reconfiguration notification to the listener MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * MBean method:
     * @return connection max open time
     */
    public Integer getJDBCMaxOpenTime() {
        return new Integer(pool.getMaxOpenTime());
    }
    /**
     * MBean method allowing to set the JDBC connection max open time
     * @param mn connection max open time in minutes
     */
    public void setJDBCMaxOpenTime(Integer mn) {
        pool.setMaxOpenTime(mn.intValue());
        String propName = DataBaseServiceImpl.MAXOPENTIME;
        ReconfiguredProp prop = new ReconfiguredProp(propName, mn.toString());
        // Send a reconfiguration notification to the listener MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * MBean method:
     * @return min size of connection pool
     */
    public Integer getJDBCMinConnPool() {
        return new Integer(pool.getPoolMin());
    }

    /**
     * MBean method allowing to set the minimum size of connection pool
     * @param min
     */
    public void setJDBCMinConnPool(Integer min) {
        logger.log(BasicLevel.DEBUG, "" + min);
        pool.setPoolMin(min.intValue());
        // Update the configuration file
        String propName = DataBaseServiceImpl.MINCONPOOL;
        ReconfiguredProp prop = new ReconfiguredProp(propName, min.toString());
        // Send a reconfiguration notification to the listener MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * MBean method:
     * @return max size of connection pool
     */
    public Integer getJDBCMaxConnPool() {
        return new Integer(pool.getPoolMax());
    }

    /**
     * MBean method allowing to set the maximum size of connection pool
     * @param max
     */
    public void setJDBCMaxConnPool(Integer max) {
        logger.log(BasicLevel.DEBUG, "" + max);
        pool.setPoolMax(max.intValue());
        // Update the configuration file
        String propName = DataBaseServiceImpl.MAXCONPOOL;
        ReconfiguredProp prop = new ReconfiguredProp(propName, max.toString());
        // Send a reconfiguration notification to the listener MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * MBean method:
     * @return test statement 
     */
    public String getJDBCTestStatement() {
        return pool.getTestStatement();
    }

    /**
     * MBean method allowing to set the SQL query for test statement
     * @param test statement 
     */
    public void setJDBCTestStatement(String test) {
        pool.setTestStatement(test);
        String propName = DataBaseServiceImpl.CONNTESTSTMT;
        ReconfiguredProp prop = new ReconfiguredProp(propName, test);
        // Send a reconfiguration notification to the listener MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }
    
    /**
     * MBean method:
     * @return max time to wait for a connection, in seconds
     */
    public Integer getJDBCMaxWaitTime() {
        return new Integer(pool.getMaxWaitTime());
    }

    /**
     * MBean method allowing to set the maximum time to wait for a connection
     * @param max
     */
    public void setJDBCMaxWaitTime(Integer max) {
        logger.log(BasicLevel.DEBUG, "" + max);
        pool.setMaxWaitTime(max.intValue());
        // Update the configuration file
        String propName = DataBaseServiceImpl.MAXWAITTIME;
        ReconfiguredProp prop = new ReconfiguredProp(propName, max.toString());
        // Send a reconfiguration notification to the listener MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * MBean method:
     * @return max nb of waiters allowed
     */
    public Integer getJDBCMaxWaiters() {
        return new Integer(pool.getMaxWaiters());
    }

    /**
     * MBean method allowing to set the maximum nb of concurrent waiters
     * for a JDBC Connection.
     * @param max
     */
    public void setJDBCMaxWaiters(Integer max) {
        logger.log(BasicLevel.DEBUG, "" + max);
        pool.setMaxWaiters(max.intValue());
        // Update the configuration file
        String propName = DataBaseServiceImpl.MAXWAITERS;
        ReconfiguredProp prop = new ReconfiguredProp(propName, max.toString());
        // Send a reconfiguration notification to the listener MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * MBean method:
     * @return sampling period
     */
    public Integer getJDBCSamplingPeriod() {
        return new Integer(pool.getSamplingPeriod());
    }

    /**
     * MBean method allowing to set the sampling period
     * @param i
     */
    public void setJDBCSamplingPeriod(Integer i) {
        logger.log(BasicLevel.DEBUG, "" + i);
        pool.setSamplingPeriod(i.intValue());
        // Update the configuration file
        String propName = DataBaseServiceImpl.SAMPLINGPERIOD;
        ReconfiguredProp prop = new ReconfiguredProp(propName, i.toString());
        // Send a reconfiguration notification to the listener MBean
        sendReconfigNotification(getSequenceNumber(), datasourceName, prop);
    }

    /**
     * Mbean Method:
     * return the type of this data source
     */
    public String getDataSourceType() {
        String result = null;
        if (getClassName().equals("oracle.jdbc.driver.OracleDriver")) result = "Oracle";
        if (getClassName().equals("interbase.interclient.Driver")) result = "InterBase";
        if (getClassName().equals("org.enhydra.instantdb.jdbc.idbDriver")) result = "Instant DB";
        if (getClassName().equals("postgresql.Driver")) result = "Postgres";     // For V. 6.x
        if (getClassName().equals("org.postgresql.Driver")) result = "Postgres"; // For V. 7.x
        return result;
    }

    /**
     * MBean Method:
     * @return the desrciption of this datasource
     */
    public String getDataSourceDescription() {
        return dsDescription;
    }

    /**
     * Management Method:
     * @param dsDesc the desrciption of this datasource
     */
    public void setDataSourceDescription(String dsDesc) {
        dsDescription = dsDesc;
    }


    /**
     * save updated configuration
     */
    public void saveConfig() { 
        sendSaveNotification(getSequenceNumber(), datasourceName);
    }
}
