/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JEntitySwitchCS.java,v 1.1.2.1 2004/02/02 09:51:22 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.util.HashMap;
import javax.ejb.EJBException;
import javax.ejb.EntityBean;
import javax.ejb.ObjectNotFoundException;
import javax.ejb.TimedObject;
import javax.ejb.Timer;
import javax.ejb.TimerService;
import javax.ejb.TransactionRolledbackLocalException;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JEntitySwitch is used internally to synchronize accesses to the entity context
 * and thus to the entity bean instance.
 * All parts common to EJBObject and EJBLocalObject should be here.
 * Different policies can be applied to manage context/instance pairs:
 * - only 1 pair (container manages the transaction isolation)
 * - 1 pair for each transaction (transaction isolation managed by DataBase)
 * - 2 pairs (1 for transactional accesses, 1 for non transaction accesses)
 * @author Philippe Durieux
 * @author Philippe Coq
 */
public class JEntitySwitchCS extends JEntitySwitch {

    /**
     * unique EntityContext
     */
    protected JEntityContext itContext = null;

    /**
     * reentrant=true if a bean instance can be accessed concurrently.
     * this is only pertinent outside transactions.
     */
    protected boolean reentrant;

    /**
     * True if a TX need this instance currently used outside tx
     * When the last release is done, we must store this instance.
     */
    protected boolean mustStore = false;

    // max time to wait when in a transaction, before checking if
    // transaction has been set rollback only (deadlock detection)
    protected static int maxtime = 3000; // 3 sec.

    /**
     * empty constructor. Object is initialized via init() because it is
     * implemented differently according to jorm mappers.
     */
    public JEntitySwitchCS() {
        lockpolicy = EntityDesc.LOCK_CONTAINER_SERIALIZED;
        txUpdates = false;      // TODO : can set this in DD.
    }

    protected void initpolicy(JEntityFactory bf) {
        reentrant = bf.isReentrant();
        lazyregister = false;
    }

    protected JEntityContext getContext4Tx(Transaction tx) {
        return itContext;
    }

    protected void setContext4Tx(Transaction tx, JEntityContext ctx) {
        itContext = ctx;
    }

    protected void removeContext4Tx(Transaction tx) {
        discarded = false;
        itContext = null;
    }

    /**
     * Map a context and its instance.
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind if not null
     * @param forced - force to take this context. (case of create)
     * @param holdit - increment count to hold it, a release will be called later.
     * @return JEntityContext actually mapped
     */
    public synchronized JEntityContext mapICtx(Transaction tx, JEntityContext bctx, boolean forced, boolean holdit) {

        // Synchronization.
        if (tx == null) {
            // ------------- IH begin -------------
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH");

            // Must wait in case of TX or if non reentrant and already busy
            while ((runningtx != null) || (! reentrant && countIH > 0)) {
                if (runningtx != null) {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH: WAIT end IT");
                } else {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH: WAIT end IH");
                }
                waiters++;
                try {
                    wait(maxtime);
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH: NOTIFIED");
                } catch (InterruptedException e) {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH: INTERRUPTED");
                } catch (Exception e) {
                    throw new EJBException("JEntitySwitch synchronization pb", e);
                } finally {
                    waiters--;
                }
            }
        } else {
            // ------------- IT begin -------------
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT");

            // Must wait in case of TX or if instance has been modified outside transaction.
            // Don't wait transactions if 1 instance per transaction.
            int waitcount = 0;
            while (inDirtyList || (runningtx != null && !tx.equals(runningtx))) {
                // instance is dirty : must write it before working on it again.
                // this is mandatory to be able to retrieve a good state in case of rollback.
                if (inDirtyList) {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT: WAIT end IH");
                    // Cannot passivate here, because must be done outside tx context
                    bf.synchronizeEntities();
                } else {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT: WAIT end IT");
                    // deadlock detection
                    blockedtx = tx;
                    if (waitcount++ > 1 && bf.isBlocked(runningtx) && bf.isBlocking(tx)) {
                        try {
                            tx.setRollbackOnly();
                        } catch (SystemException e) {
                            TraceEjb.logger.log(BasicLevel.ERROR, ident + "getICtx IT: unexpected exception setting rollbackonly");
                        }
                        TraceEjb.logger.log(BasicLevel.WARN, ident + "getICtx IT: transaction rolled back");
                        throw new TransactionRolledbackLocalException("possible deadlock");
                    }
                }
                waiters++;
                try {
                    wait(maxtime);
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT: NOTIFIED");
                } catch (InterruptedException e) {
                    if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT: INTERRUPTED");
                } catch (Exception e) {
                    throw new EJBException("JEntitySwitch synchronization pb", e);
                } finally {
                    waiters--;
                    blockedtx = null;
                }
                // If transaction has been rolledback or set rollback only, give up.
                int status = Status.STATUS_ROLLEDBACK;
                try {
                    status = tx.getStatus();
                } catch (SystemException e) {
                    TraceEjb.logger.log(BasicLevel.ERROR, ident + "getICtx IT: unexpected exception getting transaction status");
                }
                switch (status) {
                case Status.STATUS_MARKED_ROLLBACK:
                case Status.STATUS_ROLLEDBACK:
                case Status.STATUS_ROLLING_BACK:
                    TraceEjb.logger.log(BasicLevel.WARN, ident + "getICtx IT: transaction rolled back");
                    throw new TransactionRolledbackLocalException("rollback occured while waiting");
                }
            }
        }

        // Choose the context to use.
        boolean newtrans = false;
        JEntityContext jec = itContext;
        if (forced) {
            // If the new context is enforced, we must first release the older
            if (jec != null) {
                if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "new context is enforced!");
                bf.releaseJContext(jec);
                itContext = null;
                discarded = false;
            }
            jec = bctx;
            itContext = jec;
            jec.initEntityContext(this);
            newtrans = true;
        } else {
            if (jec != null) {
                // Reuse the Context for this transaction.
                // If a context was supplied, release it first.
                if (bctx != null) {
                    if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "a context was supplied!");
                    bf.releaseJContext(bctx);
                }
                jec.reuseEntityContext();
                // In case the same instance is used for all tx, must check if new one.
                if (runningtx == null) {
                    newtrans = true;
                }
            } else {
                if (bctx != null) {
                    jec = bctx;
                } else {
                    // no Context available : get one from the pool.
                    jec = (JEntityContext) bf.getJContext();
                }
                jec.initEntityContext(this);
                jec.activate(true);
                itContext = jec; // after activate
                newtrans = true;
            }
        }

        if (tx != null) {
            // Register Context now, except if no new transaction
            if (newtrans) {
                if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT: new tx, registerSynchronization");
                registerCtx(tx, jec);
            }
        } else {
            if (holdit) {
                countIH++;
                if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH count=" + countIH);
                if (shared && countIH == 1) {
                    // reload state that could have been modified by transactions.
                    jec.activate(false);
                }
            }
            if (!inDirtyList && !txUpdates) {
                inDirtyList = true;
                bf.registerEJB(this);
            }
        }

        return jec;
    }


    /**
     * Release a context/instance at end of request.
     * @param tx - transaction associated to this context
     */
    public synchronized void releaseICtx(Transaction tx) {

        if (tx == null) {
            // ------------- IH end -------------
            JEntityContext jec = itContext;
            countIH--;
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "releaseICtx IH count=" + countIH);
            if (countIH == 0) {
                if (discarded) {
                    if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "discarded!");
                    bf.releaseJContext(jec);
                    itContext = null;
                    discarded = false;
                } else {
                    if (jec.isMarkedRemoved()) {
                        if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "Marked Removed!");
                        bf.releaseJContext(jec);
                        itContext = null;
                        noLongerUsed();
                    } else if (mustStore) {
                        try {
                            jec.storeIfModified();
                        } catch (EJBException e) {
                            TraceEjb.logger.log(BasicLevel.WARN, ident + " ejbexception:" + e);
                        }
                        mustStore = false;
                    }
                }
                if (waiters > 0) {
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
                    notifyAll();
                }
            }
        } else {
            // ------------- IT end -------------
            // nothing to do now. Wait for tx completed.
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "releaseICtx IT");

        }
    }

    /**
     * try to passivate IH (called from swapper)
     * @param passivation false if only store bean state.
     * @return true if OK, false if busy.
     */
    public synchronized boolean passivateIH(boolean passivation) {

        JEntityContext jec = itContext;

        // If no context mapped or readonly, nothing to do
        if (jec == null || txUpdates) {
            return true;
        }

        // If the instance is busy, don't touch it.
        if (countIH > 0) {
            // will be stored later, when released by last thread (countIH = 0)
            // return true to remove from the dirty list, since it will be written anyway.
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " used off TX");
            mustStore = true;
            return true;
        }
        if (runningtx != null) {
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " used in TX");
            return false;
        }

        if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident);

        // Store the instance state
        // -> only if instance not discarded
        if (!discarded) {
            try {
                jec.storeIfModified();
            } catch (Exception e) {
                TraceEjb.logger.log(BasicLevel.ERROR, ident, "error while storing bean state:", e);
            }
            mustStore = false;
            // don't set dirty false here, will be done in endIH.
            // this prevent duplicates in dirty list.
        }

        // passivate this instance if required.
        if (passivation) {
            if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, "passivated: " + jec);
            jec.passivate();
            bf.releaseJContext(jec);
            itContext = null;
            discarded = false;
            removeContext4Tx(null);
            if (waiters > 0) {
                TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
                notifyAll();
            }
        }
        return true;
    }

    /**
     * Instance is ready to use for new transaction.
     */
    public synchronized void endIH() {
        inDirtyList = false;
        if (itContext == null) {
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " discarded!");
            return;
        }
        // If the instance is busy, don't touch it.
        // If no context mapped, nothing to do
        if (countIH == 0) {
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " ready again");
        } else {
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + " busy!");
        }
        if (waiters > 0) {
            TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
            notifyAll();
        }
    }

    public synchronized void notifyWriting(Transaction tx, JEntityContext bctx) {
        return;                  // NEVER
    }
}
